(***********************************************************************)
(*                                                                     *)
(*                          OcamlQuake 0.1                             *)
(*                                                                     *)
(*                         Francois Pessaux                            *)
(*                                                                     *)
(*  Copyright 1998                                                     *)
(*  This code can be re-used or modified without any restriction.      *)
(*  This version is still under development, there may be bugs and     *)
(*  inefficient stuff, and I will not be responsible of what can       *)
(*  happen to you and your environment if you use, read or do          *)
(*  whatever you want with all this... Enjoy...                        *)
(*                                                                     *)
(***********************************************************************)


open Qheader ;;
open Math ;;


type camera = {
  position : vec3_t ;
  (* Matrice reprsentant la rotation qu'a subi le joueur *)
  rotation : Matrix.matrix3x3 ;
  (* Matrice reprsentant la rotation oppose de celle qu'a subi le joueur *)
  (* Sert  transformer les objets dans le repre view-space *)
  irotation : Matrix.matrix3x3 ;
  (* Angle (en radians) de vise autour de l'axe z : direction horizontale *)
  mutable h_angle : float ;
  (* Angle (en radians) de vise autour de l'axe y : direction verticale *)
  mutable v_angle : float
  } ;;



open Matrix ;;

(* ******************
Mis  jour par draw_face. Comme ca ne change pas pour chaque vertex d'une mme
face, a nous vite de passer a en argument a la fonction
'vertex_to_camera_space' et au moins c'est extrait une fois pour toutes.
*)
(* ******************
val vertex_to_camera_space : camera -> Qheader.vertex_t -> Gr.vertex
*)
let _vectorS = ref { x = 0.0 ; y = 0.0 ; z = 0.0 } ;;
let _distS = ref 0.0 ;;
let _vectorT = ref { x = 0.0 ; y = 0.0 ; z = 0.0 } ;;
let _distT = ref 0.0 ;;



(* ******************
Transforme un vertex en camera-view-space et au passage en fait un vertex
Glide dans la foule. Synthtise s et t et les mets dans le tmu vertex.
Attention, il s'agit de s et t et non sow et tow comme l'indique le nom du
champ. Passer des permiers aux second sera fait au moment de la projection
vers le monde 2D.
*)
let vertex_to_camera_space camera ({ vx = vx ; vy = vy ; vz = vz } as vertex) =
  (* Partage de camera.position *)
  let camera_position = camera.position in
  let vx' = vx -. camera_position.x in
  let vy' = vy -. camera_position.y in
  let vz' = vz -. camera_position.z in
  (* Inlining du corps de vertex_rotate *)
  let matrix = camera.irotation in
  let vx'' = vx' *. matrix.m11 +. vy' *. matrix.m12 +. vz' *. matrix.m13 in
  let vy'' = vx' *. matrix.m21 +. vy' *. matrix.m22 +. vz' *. matrix.m23 in
  let vz'' = vx' *. matrix.m31 +. vy' *. matrix.m32 +. vz' *. matrix.m33 in
  let base_vertex =
    { Gr.x = vx'' ; Gr.y = vy'' ;
      Gr.ooz = vz'' ; (* ooz n'est jamais utilise par Glide donc on s'en *)
                      (* sert pour mmoriser la composante z une fois    *)
                      (* transforme. *)
      Gr.oow = 0.0 ; Gr.r = 0.0 ; Gr.g = 0.0 ; Gr.b = 0.0 ; Gr.a = 0.0 } in
  (* Expansion du produit scalaire en ligne pour viter l'appel de fct
  let s = (dotprod vertex !_vectorS) +. !_distS in
  let t = (dotprod vertex !_vectorT) +. !_distT in
  Et partage de vertex.vx, vertex.vy, vertex.vz
  *)
  let vertex_vx = vertex.vx in
  let vertex_vy = vertex.vy in
  let vertex_vz = vertex.vz in
  let s =
    let vs = !_vectorS in
    (vertex_vx *. vs.x +. vertex_vy *. vs.y +. vertex_vz *. vs.z) +. !_distS in
  let t =
    let vt = !_vectorT in
    (vertex_vx *. vt.x +. vertex_vy *. vt.y +. vertex_vz *. vt.z) +. !_distT in
  (* Cration du vertex final *)
  let tmu_vertex = {
    (* On ne s'en sert jamais de celui-l *)
    Gr.tmu_oow = 0.0 ;
    (* On stocke s et t ici car ces champs ne sont pas encore utiliss *)
    (* tant donn qu'on n'est pas encore dans une smantique 2D pour  *)
    (* le vertex.                                                      *)
    Gr.tmu_sow = s ;
    Gr.tmu_tow = t } in
  { Gr.base = base_vertex ;
    Gr.tmu = [|tmu_vertex|] }
;;



(* ******************
Matrice temporaire utilise pour la prparation de la camra.
*)
(* ******************
val tmp_matrix : Matrix.matrix3x3
*)
let tmp_matrix = identity_matrix () ;;



(* ******************
Met  jour les matrices de rotation de la camra en fonction des angles
trouvs par analyse des contrleurs de mouvement.
*)
(* ******************
val prepare_camera_rotation_matrices : camera -> unit
*)
let prepare_camera_rotation_matrices camera =
  (* "Vide" la matrice en y mettant une matrice identit *)
  Matrix.reset_matrix camera.rotation ;
  (* D'abord rotation horizontale (autour de z) *)
  Matrix.fill_rz_matrix camera.h_angle tmp_matrix ;
  Matrix.mult_matrices camera.rotation tmp_matrix camera.rotation ;
  (* Puis rotation verticale (autour de y) *)
  Matrix.fill_ry_matrix camera.v_angle tmp_matrix ;
  Matrix.mult_matrices camera.rotation tmp_matrix camera.rotation ;
  (* Directement la transpose pour aller plus vite, car la rotation *)
  (* inverse c'est la transpose de la matrice de rotation directe.  *)
  Matrix.transpose_matrix camera.rotation camera.irotation
;;
