(***********************************************************************)
(*                                                                     *)
(*                          OcamlQuake 0.1                             *)
(*                                                                     *)
(*                         Francois Pessaux                            *)
(*                                                                     *)
(*  Copyright 1998                                                     *)
(*  This code can be re-used or modified without any restriction.      *)
(*  This version is still under development, there may be bugs and     *)
(*  inefficient stuff, and I will not be responsible of what can       *)
(*  happen to you and your environment if you use, read or do          *)
(*  whatever you want with all this... Enjoy...                        *)
(*                                                                     *)
(***********************************************************************)


let _tmp_output = Array.create 30 Global.dummy_glide_vertex ;;
let _tmp_distances = Array.create 30 0.0 ;;


let clip_poly polygon distance_function =
  let output_size = ref 0 in
  let allin = ref true and allout = ref true in
  for i = 0 to Array.length polygon - 1 do
    let vertex = polygon.(i) in
    let d = distance_function vertex in
    if d < 0.0 then allin := false ;
    if d >= 0.0 then allout := false ;
    _tmp_distances.(i) <- d
  done ;
  if !allin then polygon
  else if !allout then [||]
  else
    let nb_vertices = Array.length polygon in
    (* Start with the last vertex as first treated vertex *)
    let v1 = ref polygon.(nb_vertices - 1) in
    let dist1 = ref _tmp_distances.(nb_vertices - 1) in
    let inside = ref (!dist1 >= 0.0) in

    for i = 0 to nb_vertices - 1 do
      let v2 = polygon.(i) in
      let dist2 = _tmp_distances.(i) in
      if !inside && dist2 >= 0.0 then
	begin
	(* Both in *)
	_tmp_output.(!output_size) <- v2 ;
	incr output_size
	end
      else
	begin
	if (not !inside) && dist2 >= 0.0 then
	  begin
	  (* Coming in *)
	  inside := true ;
	  let d = !dist1 /. (!dist1 -. dist2) in
	  let tmpbv = {
	    Gr.x = !v1.Gr.base.Gr.x +. (v2.Gr.base.Gr.x -. !v1.Gr.base.Gr.x) *. d ;
	    Gr.y = !v1.Gr.base.Gr.y +. (v2.Gr.base.Gr.y -. !v1.Gr.base.Gr.y) *. d ;
            (* Normalement il faudrait aussi les interpoler mais comme *)
            (* on ne s'en sert pas, c'est inutile de le faire.         *)
	    Gr.ooz = 0.0 ;
	    Gr.oow = !v1.Gr.base.Gr.oow +. (v2.Gr.base.Gr.oow -. !v1.Gr.base.Gr.oow) *. d ;
	    Gr.r = 0.0 ; Gr.g = 0.0 ;
	    Gr.b = 0.0 ; Gr.a = 0.0 } in
	  let tmptmu = {
	    Gr.tmu_oow = 0.0 ;  (* Ne sert jamais *)
	    Gr.tmu_sow = !v1.Gr.tmu.(0).Gr.tmu_sow +. (v2.Gr.tmu.(0).Gr.tmu_sow -. !v1.Gr.tmu.(0).Gr.tmu_sow) *. d ;
	    Gr.tmu_tow = !v1.Gr.tmu.(0).Gr.tmu_tow +. (v2.Gr.tmu.(0).Gr.tmu_tow -. !v1.Gr.tmu.(0).Gr.tmu_tow) *. d } in
	  let tmpv = { Gr.base = tmpbv ; Gr.tmu = [|tmptmu|] } in
	    (* Output vertices *)
	    _tmp_output.(!output_size) <- tmpv ;
	    incr output_size ;
            _tmp_output.(!output_size) <- v2 ;
	    incr output_size
	   end
	else
	  begin
	  if !inside && dist2 < 0.0 then
	    begin
	    (* Going out *)
	    inside := false ;
	    let d = !dist1 /. (!dist1 -. dist2) in
	    let tmpbv = {
	      Gr.x = !v1.Gr.base.Gr.x +. (v2.Gr.base.Gr.x -. !v1.Gr.base.Gr.x) *. d ;
	      Gr.y = !v1.Gr.base.Gr.y +. (v2.Gr.base.Gr.y -. !v1.Gr.base.Gr.y) *. d ;
              (* Normalement il faudrait aussi les interpoler mais comme *)
              (* on ne s'en sert pas, c'est inutile de le faire.         *)
	      Gr.ooz = 0.0 ;
	      Gr.oow = !v1.Gr.base.Gr.oow +. (v2.Gr.base.Gr.oow -. !v1.Gr.base.Gr.oow) *. d ;
	      Gr.r = 0.0 ; Gr.g = 0.0 ;
	      Gr.b = 0.0 ; Gr.a = 0.0 } in
	    let tmptmu = {
	      Gr.tmu_oow = 0.0 ;  (* Ne sert jamais *)
	      Gr.tmu_sow = !v1.Gr.tmu.(0).Gr.tmu_sow +. (v2.Gr.tmu.(0).Gr.tmu_sow -. !v1.Gr.tmu.(0).Gr.tmu_sow) *. d ;
	      Gr.tmu_tow = !v1.Gr.tmu.(0).Gr.tmu_tow +. (v2.Gr.tmu.(0).Gr.tmu_tow -. !v1.Gr.tmu.(0).Gr.tmu_tow) *. d } in
	    let tmpv = { Gr.base = tmpbv ; Gr.tmu = [|tmptmu|] } in
	     (* Output vertex *)
	     _tmp_output.(!output_size) <- tmpv ;
	     incr output_size
	     end
	  else
	    () (* Both out *)
	  end
	end ;
      v1 := polygon.(i) ;
      dist1 := _tmp_distances.(i)
    done ;
    Array.sub _tmp_output 0 !output_size
;;




(* INCLUSIVE ! *)
let border_right = 639.0 ;;
let border_top = 479.0 ;;
let border_left = 0.0 ;;
let border_bottom = 0.0 ;;



let clip_2d_poly_to_screen vertices =
 let clipped_right = clip_poly vertices (fun v -> border_right -. v.Gr.base.Gr.x) in
 let clipped_top = clip_poly clipped_right (fun v -> border_top -. v.Gr.base.Gr.y) in
 let clipped_left = clip_poly clipped_top (fun v -> v.Gr.base.Gr.x -. border_left) in
 let clipped_bottom = clip_poly clipped_left (fun v -> v.Gr.base.Gr.y -. border_bottom)
 in clipped_bottom
;;
