(***********************************************************************)
(*                                                                     *)
(*                          OcamlQuake 0.1                             *)
(*                                                                     *)
(*                         Francois Pessaux                            *)
(*                                                                     *)
(*  Copyright 1998                                                     *)
(*  This code can be re-used or modified without any restriction.      *)
(*  This version is still under development, there may be bugs and     *)
(*  inefficient stuff, and I will not be responsible of what can       *)
(*  happen to you and your environment if you use, read or do          *)
(*  whatever you want with all this... Enjoy...                        *)
(*                                                                     *)
(***********************************************************************)


open Qheader ;;


(* 3D clipping global stuff *)
(* Tableau de vertices temporaires pour eviter les allocs *)
let _tmp_vertices = Array.create 30 Global.dummy_glide_vertex ;;
let _tmp_distances = Array.create 30 0.0 ;;
let nb_tmp_vertices = ref 0 ;;


let clip_polygon3d_to_3d_plane polygon plane =
  nb_tmp_vertices := 0 ;
  let allin = ref true and allout = ref true in
  for i = 0 to Array.length polygon - 1 do
    let vertex = polygon.(i) in
    let d = Math.vertex_plane_distance plane vertex in
    if d < 0.0 then allin := false ;
    if d >= 0.0 then allout := false ;
    _tmp_distances.(i) <- d
  done ;
  if !allin then polygon
  else if !allout then [||]
  else
    let nb_vertices = Array.length polygon in
    (* Start with the last vertex as first treated vertex *)
    let v1 = ref polygon.(nb_vertices - 1) in
    let dist1 = ref _tmp_distances.(nb_vertices - 1) in
    let inside = ref (!dist1 >= 0.0) in

    for i = 0 to nb_vertices - 1 do
      let v2 = polygon.(i) in
      let dist2 = _tmp_distances.(i) in
      if !inside && dist2 >= 0.0 then
	begin
	(* Both in *)
	_tmp_vertices.(!nb_tmp_vertices) <- v2 ;
	incr nb_tmp_vertices
	end
      else
	begin
	if (not !inside) && dist2 >= 0.0 then
	  begin
	  (* Coming in *)
	  inside := true ;
	  let d = !dist1 /. (!dist1 -. dist2) in
	  let tmpbv = {
	    Gr.x = !v1.Gr.base.Gr.x +. (v2.Gr.base.Gr.x -. !v1.Gr.base.Gr.x) *. d ;
	    Gr.y = !v1.Gr.base.Gr.y +. (v2.Gr.base.Gr.y -. !v1.Gr.base.Gr.y) *. d ;
	    Gr.ooz = !v1.Gr.base.Gr.ooz +. (v2.Gr.base.Gr.ooz -. !v1.Gr.base.Gr.ooz) *. d ;
	    (* Normalement il faudrait aussi les interpoler mais comme *)
	    (* on ne s'en sert pas, c'est inutile de le faire.         *)
	    Gr.oow = 0.0 ; Gr.r = 0.0 ; Gr.g = 0.0 ; Gr.b = 0.0 ;
	    Gr.a = 0.0 } in
	  let tmptmu = {
	    Gr.tmu_oow = 0.0 ;  (* Ne sert jamais *)
	    Gr.tmu_sow = !v1.Gr.tmu.(0).Gr.tmu_sow +. (v2.Gr.tmu.(0).Gr.tmu_sow -. !v1.Gr.tmu.(0).Gr.tmu_sow) *. d ;
	    Gr.tmu_tow = !v1.Gr.tmu.(0).Gr.tmu_tow +. (v2.Gr.tmu.(0).Gr.tmu_tow -. !v1.Gr.tmu.(0).Gr.tmu_tow) *. d }  in
	  let tmpv = { Gr.base = tmpbv ; Gr.tmu = [|tmptmu|] } in
	    (* Output vertices *)
	    _tmp_vertices.(!nb_tmp_vertices) <- tmpv ;
	    incr nb_tmp_vertices ;
            _tmp_vertices.(!nb_tmp_vertices) <- v2 ;
	    incr nb_tmp_vertices
	   end
	else
	  begin
	  if !inside && dist2 < 0.0 then
	    begin
	    (* Going out *)
	    inside := false ;
	    let d = !dist1 /. (!dist1 -. dist2) in
	    let tmpbv = {
	      Gr.x = !v1.Gr.base.Gr.x +. (v2.Gr.base.Gr.x -. !v1.Gr.base.Gr.x) *. d ;
	      Gr.y = !v1.Gr.base.Gr.y +. (v2.Gr.base.Gr.y -. !v1.Gr.base.Gr.y) *. d ;
	      Gr.ooz = !v1.Gr.base.Gr.ooz +. (v2.Gr.base.Gr.ooz -. !v1.Gr.base.Gr.ooz) *. d ;
              (* Normalement il faudrait aussi les interpoler mais comme *)
              (* on ne s'en sert pas, c'est inutile de le faire.         *)
	      Gr.oow = 0.0 ; Gr.r = 0.0 ; Gr.g = 0.0 ; Gr.b = 0.0 ;
	      Gr.a = 0.0 } in
	    let tmptmu = {
	      Gr.tmu_oow = 0.0 ;  (* Ne sert jamais *)
	      Gr.tmu_sow = !v1.Gr.tmu.(0).Gr.tmu_sow +. (v2.Gr.tmu.(0).Gr.tmu_sow -. !v1.Gr.tmu.(0).Gr.tmu_sow) *. d ;
	      Gr.tmu_tow = !v1.Gr.tmu.(0).Gr.tmu_tow +. (v2.Gr.tmu.(0).Gr.tmu_tow -. !v1.Gr.tmu.(0).Gr.tmu_tow) *. d } in
	    let tmpv = { Gr.base = tmpbv ; Gr.tmu = [|tmptmu|] } in
	     (* Output vertex *)
	     _tmp_vertices.(!nb_tmp_vertices) <- tmpv ;
	     incr nb_tmp_vertices
	     end
	  else
	    () (* Both out *)
	  end
	end ;
      v1 := polygon.(i) ;
      dist1 := _tmp_distances.(i)
    done ;
    Array.sub _tmp_vertices 0 !nb_tmp_vertices
;;
