(***********************************************************************)
(*                                                                     *)
(*                          OcamlQuake 0.1                             *)
(*                                                                     *)
(*                         Francois Pessaux                            *)
(*                                                                     *)
(*  Copyright 1998                                                     *)
(*  This code can be re-used or modified without any restriction.      *)
(*  This version is still under development, there may be bugs and     *)
(*  inefficient stuff, and I will not be responsible of what can       *)
(*  happen to you and your environment if you use, read or do          *)
(*  whatever you want with all this... Enjoy...                        *)
(*                                                                     *)
(***********************************************************************)


open Qheader ;;


let vec3_dotprod { x = a1 ; y = a2 ; z = a3 } { x = b1 ; y = b2 ; z = b3 } =
  a1 *. b1 +. a2 *. b2 +. a3 *. b3
;;


(*
Code from modifications of
 "A Note on Quake BSP Algorithms by Kekoa Proudfoot"
http://www-graphics.stanford.edu/~kekoa/bsp.html
*)
let rec ray_intersect node_index p d min max =
  if node_index < 0 then         (* node_index represente une feuille *)
    if node_index = -1 then max else min
  else
    begin
    let node = !Global.clipnodes.(node_index) in
    let plane = !Global.planes.(node.Qheader.planenum) in
    (* Attention, pour les clipnodes, les normales des plans sont *)
    (* inversees et pointent vers l'exterieur. Donc le calcul de  *)
    (* distance est egalement inverse.                            *)
    let plane_normal = plane.normal in
    let dist = plane.dist -.
               (plane_normal.x *. p.Qheader.x +.
		plane_normal.y *. p.Qheader.y +.
		plane_normal.z *. p.Qheader.z) in
    let nearer = if dist <= 0.0 then node.cfront else node.cback in
    let farther = if dist <= 0.0 then node.cback else node.cfront in
    let scale = vec3_dotprod plane.normal d in
    if scale = 0.0 then ray_intersect nearer p d min max
    else
      begin
      let dist = dist /. scale in
      if dist > max || dist < 0.0 then
        ray_intersect nearer p d min max
      else
	if dist < min then
          ray_intersect farther p d min max
	else
	  begin
	  let min = ray_intersect nearer p d min dist in
	  if min <> dist then min
          else ray_intersect farther p d dist max
          end
      end
    end
;;



let exists_collision p0 p1 node_index =
  let d = {
    Qheader.x = p1.Qheader.x -. p0.Qheader.x ;
    Qheader.y = p1.Qheader.y -. p0.Qheader.y ;
    Qheader.z = p1.Qheader.z -. p0.Qheader.z } in
  let tmp = ray_intersect node_index p0 d 0.0 1.0 in
  tmp <> 1.0
;;
