(***********************************************************************)
(*                                                                     *)
(*                          OcamlQuake 0.1                             *)
(*                                                                     *)
(*                         Francois Pessaux                            *)
(*                                                                     *)
(*  Copyright 1998                                                     *)
(*  This code can be re-used or modified without any restriction.      *)
(*  This version is still under development, there may be bugs and     *)
(*  inefficient stuff, and I will not be responsible of what can       *)
(*  happen to you and your environment if you use, read or do          *)
(*  whatever you want with all this... Enjoy...                        *)
(*                                                                     *)
(***********************************************************************)


open Qheader ;;
open Math ;;
open Camera ;;
open Frustum ;;

let numentities = ref 0 ;;
let entities = ref ([||] : entity_t array) ;;

let nummodels = ref 0 ;;
let models = ref ([||] : model_t array) ;;

let numnodes = ref 0 ;;
let nodes = ref ([||] : node_t array) ;;

let numsurfaces = ref 0 ;;
let surfaces = ref ([||] : surface_t array) ;;

let numplanes = ref 0 ;;
let planes = ref ([||] : plane_t array) ;;

let numleaves = ref 0 ;;
let leaves = ref ([||] : dleaf_t array) ;;

let numvisilists = ref 0 ;;
let visilists = ref "" ;;

let numlfaces = ref 0 ;;
let lfaces = ref ([||] : int array) ;;

let numfaces = ref 0 ;;
let faces = ref ([||] : face_t array) ;;

let clipnodes = ref ([||] : clipnode_t array) ;;
let numclipnodes = ref 0 ;;

let numledges = ref 0 ;;
let ledges = ref ([||] : int array) ;;

let numedges = ref 0 ;;
let edges = ref ([||] : edge_t array) ;;

let numvertices = ref 0 ;;
let vertices = ref ([||] : vertex_t array) ;;

let numsurfaces = ref 0 ;;
let surfaces = ref ([||] : surface_t array) ;;

let nummiptexs = ref 0 ;;
let miptexs = ref ([||] : Gr.tex_info array) ;;
let mipaddrs = ref ([||] : int array) ;;
(* Tableau donnant le scaling de chaque texture pour faire le lien *)
(* vec le systme de coordonnes de Glide.                         *)
let texture_scaling = ref ([||] : float array) ;;

(* Tableau disant si chaque est considre comme visible *)
let visible_leaves = ref ([||] : bool array) ;;
let visible_nodes = ref ([||] : bool array) ;;
let visible_faces = ref ([||] : bool array) ;;


let camera = {
  position = { x = 544.0 ; y = 288.0 ; z = 32.0 } ;
  rotation = Matrix.identity_matrix () ;
  irotation = Matrix.identity_matrix () ;
  h_angle = 0.0 ;
  v_angle = 0.0
  } ;;

let frustum =
  (* Plan bornant la vue vers le haut (1/2 fov vertical = 45) *)
  let up =
    { normal = vec3_normalize { x = 1.0 ; y = 0.0 ; z = -. 1.0 } ;
      dist = 0.0 ; ptype = 6 } in
  (* Plan bornant la vue vers le bas (1/2 fov vertical = 45) *)
  let down =
    { normal = vec3_normalize { x = 1.0 ; y = 0.0 ; z = 1.0 } ;
      dist = 0.0 ; ptype = 6 } in
  (* Plan bornant la vue sur la gauche (1/2 fov horizontal = 45) *)
  let left =
    { normal = vec3_normalize { x = 1.0 ; y = -. 1.0 ; z = 0.0 } ;
      dist = 0.0 ; ptype = 6 } in
  (* Plan bornant la vue sur la droite (1/2 fov horizontal = 45) *)
  let right =
    { normal = vec3_normalize { x = 1.0 ; y = 1.0 ; z = 0.0 } ;
      dist = 0.0 ; ptype = 6 } in
  { up = up ; down = down ; left = left ; right = right }
;;



let dummy_glide_base_vertex =
  { Gr.x = 0.0 ; Gr.y = 0.0 ; Gr.ooz = 0.0 ; Gr.oow = 0.0 ;
    Gr.r = 0.0 ; Gr.g = 0.0 ; Gr.b = 0.0 ; Gr.a = 0.0 }
;;

let dummy_glide_vertex =
  { Gr.base = dummy_glide_base_vertex ; Gr.tmu = [||] }
;;

let dummy_vec3t = { x = 0.0 ; y = 0.0 ; z = 0.0 } ;;

let dummy_model = {
  bound = { min = dummy_vec3t ; max = dummy_vec3t } ; origin = dummy_vec3t ;
  node_id0 = -1 ; node_id1 = -1 ; node_id2 = -1 ; node_id3 = -1 ;
  numleafs = -1 ; face_id = -1 ; face_num = -1 }
;;



(* Boolen disant si la dtection de collision doit tre active ou non *)
let coldet_active = ref true ;;
