/***********************************************************************/
/*                                                                     */
/*                          OcamlQuake 0.1                             */
/*                                                                     */
/*                         Francois Pessaux                            */
/*                                                                     */
/*  Copyright 1998                                                     */
/*  This code can be re-used or modified without any restriction.      */
/*  This version is still under development, there may be bugs and     */
/*  inefficient stuff, and I will not be responsible of what can       */
/*  happen to you and your environment if you use, read or do          */
/*  whatever you want with all this... Enjoy...                        */
/*                                                                     */
/***********************************************************************/


#include <stdio.h>
#include <X11/Xlib.h>

#include <mlvalues.h>
#include <alloc.h>
#include <memory.h>
#include <fail.h>


#define LAYOUT_WIDTH 1280
#define LAYOUT_HEIGHT 1024
#define LAYOUT_CENTER_X ((LAYOUT_WIDTH)/2)
#define LAYOUT_CENTER_Y ((LAYOUT_HEIGHT)/2)

Display *display ;
Window win ;


value init_controler ()
{
  GC gc ;
  XGCValues gcv ;
  unsigned long	vmask ;
  XSetWindowAttributes xswat ;


  display = XOpenDisplay ("") ;
  /* To prevent the WM to add borders and reparent the widget */
  xswat.override_redirect = True;
  vmask = CWOverrideRedirect;
  win = XCreateWindow  (display, DefaultRootWindow (display), 0, 0,
			LAYOUT_WIDTH, LAYOUT_HEIGHT, 0, 0, InputOnly, NULL,
			vmask, &xswat) ;
  gc = XCreateGC (display, DefaultRootWindow (display), 0L, &gcv) ;
  /* PointerMotionHintMask : Mouse events are broadcasted only when the one */
  /* in the queue is taken (to avoid bufferisation) */
  XSelectInput (display, win,
		KeyPressMask | PointerMotionMask | PointerMotionHintMask |
		ButtonPressMask | ButtonReleaseMask | KeyReleaseMask) ;
  /* Disables autorepeat for keyboard. */
  XAutoRepeatOff (display) ;
  /* Displays the window */
  XMapWindow (display, win) ;
  /* Set the focus to the window */
  XSetInputFocus (display, win, RevertToParent, CurrentTime) ;
  /* Center the mouse on the screen for the first time */
  XWarpPointer (display, None, win, 0, 0, 0, 0, LAYOUT_CENTER_X,
		LAYOUT_CENTER_Y) ;
  XFlush (display) ;
  return (Val_unit) ;
}



value close_controler ()
{
  XAutoRepeatOn (display) ;
  /* Removes the window */
  /* XUnmapWindow (display, win) ; */ 
  XDestroyWindow (display, win) ;
  XFlush (display) ;
  return (Val_unit) ;
}



value pending_event ()
{
  if (XPending (display)) return Val_int (1) ;
  else return (Val_int(0)) ;
}



value flush_event ()
{
  XFlush (display) ;
  return (Val_unit) ;
}


/*
type event =
  | Mousemove of int * int
  | Keypress of char
  | ButtonPress of int
  | ButtonRelease of int
  | Keyrelease of int
 */
#define MOUSEMOVE_TAG 0
#define KEYPRESS_TAG 1
#define BUTTONPRESS_TAG 2
#define BUTTONRELEASE_TAG 3
#define KEYRELEASE_TAG 4

#define MOUSEMOVE_ARITY 2
#define KEYPRESS_ARITY 1
#define BUTTONPRESS_ARITY 1
#define BUTTONRELEASE_ARITY 1
#define KEYRELEASE_ARITY 1

value wait_event ()
{
  XEvent ev ;
  value block ;
  
  XNextEvent (display, &ev) ;
  switch (ev.type) {

  case MotionNotify:
    block = alloc_small (MOUSEMOVE_ARITY, MOUSEMOVE_TAG) ;
    /* pointer x, y coordinates in event window */
    /* Use x_root, y_root for coordinates relative to root */
    Field (block, 0) = Val_int (ev.xmotion.x) ;
    Field (block, 1) = Val_int (ev.xmotion.y) ;
    break ;

  case KeyPress:
    block = alloc_small (KEYPRESS_ARITY, KEYPRESS_TAG) ;
    Field (block, 0) = Val_int (ev.xkey.keycode) ;
    break ;

  case KeyRelease:
    block = alloc_small (KEYRELEASE_ARITY, KEYRELEASE_TAG) ;
    Field (block, 0) = Val_int (ev.xkey.keycode) ;
    break ;

  case ButtonPress:
    block = alloc_small (BUTTONPRESS_ARITY, BUTTONPRESS_TAG) ;
    Field (block, 0) = Val_int (ev.xbutton.button) ;
    break ;

  case ButtonRelease:
    block = alloc_small (BUTTONRELEASE_ARITY, BUTTONRELEASE_TAG) ;
    Field (block, 0) = Val_int (ev.xbutton.button) ;
    break ;

  default:
    failwith ("Controler 'wait_event' : unexpected event") ;
    break ;
  }

  return (block) ;
}



#define NONE_TAG 0
#define SOME_TAG 1
/* CAMLlocal1 (block), CAMLlocal1 (option) ??? */
/* CAMLreturn (option) ??? */

/* Dummy stuff for XQueryPointer */
Window wdummy ;
int idummy ;

value check_event ()
{
  XEvent ev ;
  value block, option ;
  int x, y ;

  if (XCheckWindowEvent (display, win,
			 KeyPressMask | KeyReleaseMask | PointerMotionMask |
			 ButtonPressMask | ButtonReleaseMask, &ev)) {
    switch (ev.type) {

    case MotionNotify:
      /* Recover the most recent mouse position */
      XQueryPointer (display, win, &wdummy, &wdummy, &idummy, &idummy,
		    &x, &y, &idummy) ;
      /* Recenter mouse if not already on center */
      if ((x != LAYOUT_CENTER_X) || (y != LAYOUT_CENTER_Y))
	XWarpPointer (display, None, None, 0, 0, 0, 0,
		      LAYOUT_CENTER_X - x, LAYOUT_CENTER_Y - y) ;
      block = alloc_small (MOUSEMOVE_ARITY, MOUSEMOVE_TAG) ;
      /* pointer x, y coordinates in event window */
      /* Use x_root, y_root for coordinates relative to root */
      Field (block, 0) = Val_int (x - LAYOUT_CENTER_X) ;
      Field (block, 1) = Val_int (y - LAYOUT_CENTER_Y) ;
      break ;

    case KeyPress:
      block = alloc_small (KEYPRESS_ARITY, KEYPRESS_TAG) ;
      Field (block, 0) = Val_int (ev.xkey.keycode) ;
      break ;

    case KeyRelease:
      block = alloc_small (KEYRELEASE_ARITY, KEYRELEASE_TAG) ;
      Field (block, 0) = Val_int (ev.xkey.keycode) ;
      break ;

    case ButtonPress:
      block = alloc_small (BUTTONPRESS_ARITY, BUTTONPRESS_TAG) ;
      Field (block, 0) = Val_int (ev.xbutton.button) ;
      break ;

    case ButtonRelease:
      block = alloc_small (BUTTONRELEASE_ARITY, BUTTONRELEASE_TAG) ;
      Field (block, 0) = Val_int (ev.xbutton.button) ;
      break ;

    default:
      failwith ("Controler 'check_event' : unexpected event") ;
      break ;
    }

    option = alloc_small (1, SOME_TAG) ;
    Field (option, 0) = block ;
    return (option) ;
  }
  else return (Val_int (NONE_TAG)) ;      /* None */
}
