(***********************************************************************)
(*                                                                     *)
(*                          OcamlQuake 0.1                             *)
(*                                                                     *)
(*                         Francois Pessaux                            *)
(*                                                                     *)
(*  Copyright 1998                                                     *)
(*  This code can be re-used or modified without any restriction.      *)
(*  This version is still under development, there may be bugs and     *)
(*  inefficient stuff, and I will not be responsible of what can       *)
(*  happen to you and your environment if you use, read or do          *)
(*  whatever you want with all this... Enjoy...                        *)
(*                                                                     *)
(***********************************************************************)


open Math ;;
open Camera ;;


let start_time = ref 0.0 ;;
let end_time = ref 0.0 ;;
let frame_count = ref 0 ;;
let current_key = ref 0 ;;
(* Index du modle courant  afficher. Par dfaut, celui du "monde". *)
let current_model_index = ref 0 ;;
let current_model = ref Global.dummy_model ;;
let old_camera_position =
  { Qheader.x = 544.0; Qheader.y = 288.0 ; Qheader.z = 32.0 } ;;
(* Flags pour savoir quel mnage faire quand on quitte *)
let glide_is_on = ref false ;;
let win_is_open = ref false ;;
let controler_is_on = ref false ;;


let restore_old_camera_position () =
  Global.camera.position.Qheader.x <- old_camera_position.Qheader.x ;
  Global.camera.position.Qheader.y <- old_camera_position.Qheader.y ;
  Global.camera.position.Qheader.z <- old_camera_position.Qheader.z
;;


let update_camera_old_position () =
  old_camera_position.Qheader.x <- Global.camera.position.Qheader.x ;
  old_camera_position.Qheader.y <- Global.camera.position.Qheader.y ;
  old_camera_position.Qheader.z <- Global.camera.position.Qheader.z
;;



let find_initial_camera_position () =
  try
    for i = 0 to !Global.numentities - 1 do
      match !Global.entities.(i) with
       | Qheader.Info_player_start position ->
	   Global.camera.position.Qheader.x <- position.Qheader.x ;
	   Global.camera.position.Qheader.y <- position.Qheader.y ;
	   Global.camera.position.Qheader.z <- position.Qheader.z ;
	   update_camera_old_position () ;
	   raise End_of_file
       | _ -> ()
    done ;
    Printf.fprintf stderr "Warning not start position found\n"
  with End_of_file -> ()
;;



let clean_exit () =
  (* On ferme seulement ce qui a t ouvert, sinon a fait core dump. *)
  if !win_is_open then Gr.sst_win_close () ;
  if !glide_is_on then Gr.glide_shutdown () ;
  if !controler_is_on then Inputcontrol.close_controler () ;
  end_time := Unix.gettimeofday () ;
  Printf.fprintf stderr "Frame rate : %f f/s\n"
    (float !frame_count /. (!end_time -. !start_time)) ;
  flush stderr ;
  exit 0
;;



let handle_mouse mouse_x mouse_y =
  (* Vise horizontale *)
  let angle = -. (float mouse_x) /. 5.0 in
  Global.camera.h_angle <- Global.camera.h_angle +. (rad angle) ;
  (* Vise verticale *)
  let angle = -. (float mouse_y) /. 5.0 in
  Global.camera.v_angle <- Global.camera.v_angle +. (rad angle)
;;



let handle_key = function
  | 97 ->  (* Home : move forward *)
      let tmp = { Qheader.x = 20.0 ; Qheader.y = 0.0; Qheader.z = 0.0 } in
      let tmp' = vec3_rotate tmp Global.camera.rotation in
      Global.camera.position.Qheader.x <-
             Global.camera.position.Qheader.x +. tmp'.Qheader.x ;
      Global.camera.position.Qheader.y <-
             Global.camera.position.Qheader.y +. tmp'.Qheader.y ;
      Global.camera.position.Qheader.z <-
             Global.camera.position.Qheader.z +. tmp'.Qheader.z
  | 103 -> (* End : move backward *)
      let tmp = { Qheader.x = -.20.0 ; Qheader.y = 0.0; Qheader.z = 0.0 } in
      let tmp' = vec3_rotate tmp Global.camera.rotation in
      Global.camera.position.Qheader.x <-
      Global.camera.position.Qheader.x +. tmp'.Qheader.x ;
      Global.camera.position.Qheader.y <-
      Global.camera.position.Qheader.y +. tmp'.Qheader.y ;
      Global.camera.position.Qheader.z <-
      Global.camera.position.Qheader.z +. tmp'.Qheader.z
  | 82 -> (* Minus : go down *)
      let tmp = { Qheader.x = 0.0 ; Qheader.y = 0.0; Qheader.z = -20.0 } in
      let tmp' = vec3_rotate tmp Global.camera.rotation in
      Global.camera.position.Qheader.x <-
             Global.camera.position.Qheader.x +. tmp'.Qheader.x ;
      Global.camera.position.Qheader.y <-
             Global.camera.position.Qheader.y +. tmp'.Qheader.y ;
      Global.camera.position.Qheader.z <-
             Global.camera.position.Qheader.z +. tmp'.Qheader.z
  | 86 -> (* Minus : go up *)
      let tmp = { Qheader.x = 0.0 ; Qheader.y = 0.0; Qheader.z = 20.0 } in
      let tmp' = vec3_rotate tmp Global.camera.rotation in
      Global.camera.position.Qheader.x <-
             Global.camera.position.Qheader.x +. tmp'.Qheader.x ;
      Global.camera.position.Qheader.y <-
             Global.camera.position.Qheader.y +. tmp'.Qheader.y ;
      Global.camera.position.Qheader.z <-
             Global.camera.position.Qheader.z +. tmp'.Qheader.z
  | 107 -> (* del : strafe left *)
      let tmp = { Qheader.x = 0.0 ; Qheader.y = 20.0; Qheader.z = 0.0 } in
      let tmp' = vec3_rotate tmp Global.camera.rotation in
      Global.camera.position.Qheader.x <-
             Global.camera.position.Qheader.x +. tmp'.Qheader.x ;
      Global.camera.position.Qheader.y <-
             Global.camera.position.Qheader.y +. tmp'.Qheader.y ;
      Global.camera.position.Qheader.z <-
             Global.camera.position.Qheader.z +. tmp'.Qheader.z
  | 105 -> (* pgdn : strafe right *)
      let tmp = { Qheader.x = 0.0 ; Qheader.y = -.20.0; Qheader.z = 0.0 } in
      let tmp' = vec3_rotate tmp Global.camera.rotation in
      Global.camera.position.Qheader.x <-
             Global.camera.position.Qheader.x +. tmp'.Qheader.x ;
      Global.camera.position.Qheader.y <-
             Global.camera.position.Qheader.y +. tmp'.Qheader.y ;
      Global.camera.position.Qheader.z <-
             Global.camera.position.Qheader.z +. tmp'.Qheader.z
  | 99 -> (* pgup : turn right *)
      Global.camera.h_angle <- Global.camera.h_angle +. (rad (-. 10.0))
  | 106 -> (* insert : turn left *)
      Global.camera.h_angle <- Global.camera.h_angle +. (rad 10.0)
  | 54 -> (* c : clipping on / off *)
      Global.coldet_active := not !Global.coldet_active
  | 9 -> (* Escape : quit *)
      clean_exit ()
  | _ -> ()
;;



let download_textures () =
  Global.mipaddrs := Array.create !Global.nummiptexs 0 ;
  let mem_ptr = ref (Gr.tex_min_address 0) in
  let max_mem_ptr = Gr.tex_max_address 0 in
  for i = 0 to !Global.nummiptexs - 1 do
    let current_miptex = !Global.miptexs.(i) in
    let requiered_mem =
      Gr.tex_texture_mem_required Gr.MIPMAPLEVEL_MASK_BOTH current_miptex in
    if !mem_ptr + requiered_mem > max_mem_ptr then
      failwith "Not enough room on chip for textures" ;
    Gr.tex_download_mipmap 0 !mem_ptr Gr.MIPMAPLEVEL_MASK_BOTH current_miptex ;
    !Global.mipaddrs.(i) <- !mem_ptr ;
    mem_ptr := !mem_ptr + requiered_mem
    done
;;


let download_palette () =
  let in_channel = open_in_bin "palette.lmp" in
  (* Format pour Glide: 0xFFrrggbb *)
  (* Format dans le fichier: 0xrrggbb *)
  (* Mais il faut penser  stocker les poids forts en premier *)
  let palette = String.create (256 * 4) in
  for i = 0 to 255 do
    palette.[i * 4 + 3] <- Char.chr 0xFF ;
    palette.[i * 4 + 2] <- Char.chr (Io.r_byte in_channel) ;
    palette.[i * 4 + 1] <- Char.chr (Io.r_byte in_channel) ;
    palette.[i * 4] <- Char.chr (Io.r_byte in_channel)
    done ;
  close_in in_channel ;
  Gr.tex_download_table 0 Gr.TEXTABLE_PALETTE palette
;;



let main () =
  Arg.parse [ ("-m", Arg.Int (fun id -> current_model_index := id),
	      "Specify the model to render") ]
    (fun filename -> Readbspfile.load_file filename) "Bla bla" ;
  (* Dcide quel modle on va rendre. On vrifie qu'il existe bien d'abord. *)
  if !current_model_index >= !Global.nummodels then
    failwith ("Invalid model number to render: #" ^
	      (string_of_int !current_model_index)) ;
  current_model := !Global.models.(!current_model_index) ;
  (* Recherche la position de dpart du joueur dans le tableau. *)
  find_initial_camera_position () ;
  (* Stuff d'initialisation graphique et de contrle de mouvements. *)
  Inputcontrol.init_controler () ;
  controler_is_on := true ;
  Gr.glide_init () ;
  glide_is_on := true ;
  Gr.sst_select 0 ;
  Gr.sst_win_open 0 Gr.RESOLUTION_640x480 Gr.REFRESH_60Hz Gr.COLORFORMAT_RGB
                  Gr.ORIGIN_LOWER_LEFT 2 0 ;
  win_is_open := true ;
  Gr.cull_mode Gr.CULL_DISABLE ;

  download_textures () ;
  download_palette () ;

  Gr.color_combine Gr.COMBINE_FUNCTION_SCALE_OTHER
    Gr.COMBINE_FACTOR_ONE Gr.COMBINE_LOCAL_CONSTANT Gr.COMBINE_OTHER_TEXTURE
    false ;
  Gr.tex_combine 0 Gr.COMBINE_FUNCTION_LOCAL Gr.COMBINE_FACTOR_ONE
    Gr.COMBINE_FUNCTION_LOCAL Gr.COMBINE_FACTOR_ONE false false ;
  Gr.tex_mipmap_mode 0 Gr.MIPMAP_DISABLE false ;
  Gr.tex_filter_mode 0 Gr.TEXTUREFILTER_BILINEAR Gr.TEXTUREFILTER_BILINEAR ;
  Gr.tex_clamp_mode 0 Gr.TEXTURECLAMP_WRAP Gr.TEXTURECLAMP_WRAP ;

  start_time := Unix.gettimeofday () ;
  frame_count := 1 ;
  (* Premire frame *)
  Render.render !current_model Global.camera ;
  Gr.buffer_swap 0 ;
  while true do
    (* On dessine seulement si les autres frames ont t termines *)
    if Gr.buffer_num_pending () <= 1 then
      begin
      (match Inputcontrol.check_event () with
       | None -> ()
       | Some (Inputcontrol.KeyPress key) -> current_key := key
       | Some (Inputcontrol.KeyRelease _) -> current_key := 0
       | Some (Inputcontrol.Mousemove (mouse_x, mouse_y)) ->
	   handle_mouse mouse_x mouse_y
       | _ -> ()) ;
      (* Met  jour les matrices de rotation  partir des angles *)
      prepare_camera_rotation_matrices Global.camera ;
      (* Maintenant que les matrices de rotation sont correctes, *)
      (* on peut traiter les dplacements.                       *)
      handle_key !current_key ;
      if !current_key <> 0 then
	begin
        (* Vrifier les collisions seulement si on a chang de place *)
	if !Global.coldet_active &&
	   Collision.exists_collision
	          old_camera_position Global.camera.position
	          !current_model.Qheader.node_id1
	then restore_old_camera_position ()
	else update_camera_old_position ()
	end ;
      Render.render !current_model Global.camera ;
      incr frame_count ;
      Gr.buffer_swap 0
      end
  done
;;


try main ()
with whatever ->
  Printf.fprintf stderr "Uncaught exception: %s\n"
                 (Printexc.to_string whatever) ;
  clean_exit ()
;;
