(***********************************************************************)
(*                                                                     *)
(*                          OcamlQuake 0.1                             *)
(*                                                                     *)
(*                         Francois Pessaux                            *)
(*                                                                     *)
(*  Copyright 1998                                                     *)
(*  This code can be re-used or modified without any restriction.      *)
(*  This version is still under development, there may be bugs and     *)
(*  inefficient stuff, and I will not be responsible of what can       *)
(*  happen to you and your environment if you use, read or do          *)
(*  whatever you want with all this... Enjoy...                        *)
(*                                                                     *)
(***********************************************************************)


type matrix3x3 = {
  mutable m11 : float ;
  mutable m12 : float ;
  mutable m13 : float ;
  mutable m21 : float ;
  mutable m22 : float ;
  mutable m23 : float ;
  mutable m31 : float ;
  mutable m32 : float ;
  mutable m33 : float
  } ;;



(*
NOTE SUR LES MATRICES DE ROTATION:
  Sens de rotation directe: Mcamera *. Mrotation_directe
  Sens de rotation inverse: Mrotation_inverse * Mcamera
  Mrotation_inverse = transposee de Mrotation_directe
*)



(* ******************
Gnre une matrice identit
*)
(* ******************
val identity_matrix : unit -> matrix3x3
*)
let identity_matrix () = {
  m11 = 1.0 ; m12 = 0.0 ; m13 = 0.0 ;
  m21 = 0.0 ; m22 = 1.0 ; m23 = 0.0 ;
  m31 = 0.0 ; m32 = 0.0 ; m33 = 1.0 }
;;



(* ******************
Multiplication de matrice en place  3 operandes: left * right -> result.
Peut-tre utilis avec overlap (cd avec une des source tant en mme temps
destination.
*)
(* ******************
val mult_matrices : matrix3x3 -> matrix3x3 -> matrix3x3 -> unit
*)
let mult_matrices left right result =
  let m11l = left.m11 in let m11r = right.m11 in
  let m12l = left.m12 in let m12r = right.m12 in
  let m13l = left.m13 in let m13r = right.m13 in
  let m21l = left.m21 in let m21r = right.m21 in
  let m22l = left.m22 in let m22r = right.m22 in
  let m23l = left.m23 in let m23r = right.m23 in
  let m31l = left.m31 in let m31r = right.m31 in
  let m32l = left.m32 in let m32r = right.m32 in
  let m33l = left.m33 in let m33r = right.m33 in
  result.m11 <- m11l *. m11r +. m12l *. m21r +. m13l *. m31r ;
  result.m12 <- m11l *. m12r +. m12l *. m22r +. m13l *. m32r ;
  result.m13 <- m11l *. m13r +. m12l *. m23r +. m13l *. m33r ;
  result.m21 <- m21l *. m11r +. m22l *. m21r +. m23l *. m31r ;
  result.m22 <- m21l *. m12r +. m22l *. m22r +. m23l *. m32r ;
  result.m23 <- m21l *. m13r +. m22l *. m23r +. m23l *. m33r ;
  result.m31 <- m31l *. m11r +. m32l *. m21r +. m33l *. m31r ;
  result.m32 <- m31l *. m12r +. m32l *. m22r +. m33l *. m32r ;
  result.m33 <- m31l *. m13r +. m32l *. m23r +. m33l *. m33r
;;



(* ******************
Initialise une matrice en place pour en faire la matrice dcrivant la rotation
d'angle 'a' autour de l'axe des y (repre right-handed, cd mathmatique
conventionnel).
*)
(* ******************
val fill_ry_matrix : float -> matrix3x3 -> unit
*)
let fill_ry_matrix a matrix =
  matrix.m11 <- cos a ;
  matrix.m12 <- 0.0 ;
  matrix.m13 <- sin a ;
  matrix.m21 <- 0.0 ;
  matrix.m22 <- 1.0 ;
  matrix.m23 <- 0.0 ;
  matrix.m31 <- -. sin a ;
  matrix.m32 <- 0.0 ;
  matrix.m33 <- cos a
;;



(* ******************
Initialise une matrice en place pour en faire la matrice dcrivant la rotation
d'angle 'a' autour de l'axe des z (repre right-handed, cd mathmatique
conventionnel).
*)
(* ******************
val fill_rz_matrix : float -> matrix3x3 -> unit
*)
let fill_rz_matrix a matrix =
  matrix.m11 <- cos a ;
  matrix.m12 <- -. sin a ;
  matrix.m13 <- 0.0 ;
  matrix.m21 <- sin a ;
  matrix.m22 <- cos a ;
  matrix.m23 <- 0.0 ;
  matrix.m31 <- 0.0 ;
  matrix.m32 <- 0.0 ;
  matrix.m33 <- 1.0
;;


(* ******************
Initialise une matrice en place pour en faire la matrice dcrivant la rotation
d'angle 'a' autour de l'axe des x (repre right-handed, cd mathmatique
conventionnel).
*)
(* ******************
val fill_rx_matrix : float -> matrix3x3 -> unit
*)
let fill_rx_matrix a matrix =
  matrix.m11 <- 1.0 ;
  matrix.m12 <- 0.0 ;
  matrix.m13 <- 0.0 ;
  matrix.m21 <- 0.0 ;
  matrix.m22 <- cos a ;
  matrix.m23 <- -. sin a ;
  matrix.m31 <- 0.0 ;
  matrix.m32 <- sin a ;
  matrix.m33 <- cos a
;;



(* ******************
Remplit en place une matrice pour en faire une matrice identit. Evite ainsi
de l'allocation.
*)
(* ******************
val reset_matrix : matrix3x3 -> unit
*)
let reset_matrix matrix =
  matrix.m11 <- 1.0 ;
  matrix.m12 <- 0.0 ;
  matrix.m13 <- 0.0 ;
  matrix.m21 <- 0.0 ;
  matrix.m22 <- 1.0 ;
  matrix.m23 <- 0.0 ;
  matrix.m31 <- 0.0 ;
  matrix.m32 <- 0.0 ;
  matrix.m33 <- 1.0
;;



(* ******************
Transposition de matrice a 2 oprandes: src(-1) -> dest.
Attention, ne peut pas tre utilise avec overlap.
*)
(* ******************
val transpose_matrix : matrix3x3 -> matrix3x3 -> unit
*)
let transpose_matrix src dest =
  dest.m11 <- src.m11 ;
  dest.m12 <- src.m21 ;
  dest.m13 <- src.m31 ;
  dest.m21 <- src.m12 ;
  dest.m22 <- src.m22 ;
  dest.m23 <- src.m32 ;
  dest.m31 <- src.m13 ;
  dest.m32 <- src.m23 ;
  dest.m33 <- src.m33
;;
