(***********************************************************************)
(*                                                                     *)
(*                          OcamlQuake 0.1                             *)
(*                                                                     *)
(*                         Francois Pessaux                            *)
(*                                                                     *)
(*  Copyright 1998                                                     *)
(*  This code can be re-used or modified without any restriction.      *)
(*  This version is still under development, there may be bugs and     *)
(*  inefficient stuff, and I will not be responsible of what can       *)
(*  happen to you and your environment if you use, read or do          *)
(*  whatever you want with all this... Enjoy...                        *)
(*                                                                     *)
(***********************************************************************)


open Qheader ;;
open Math ;;



(* let walk_bsp_tree_fb node_index =
   if player is front then front / back
   else back / front
;;

let walk_bsp_tree_bf node_index =
  if player is front then back / front
  else front / back
;; *)


(* Teste si un index est un index de feuille *)
let is_leaf_index index = index land 0x8000 <> 0 ;;
(* Teste si un index est un index de noeud *)
let is_node_index index = index land 0x8000 = 0 ;;



(* ******************
Recherche dans quelle feuille du BSP se trouve le point argument
*)
(* ******************
val find_lying_leaf : Qheader.vec3_t -> int
*)
let find_lying_leaf model point =
  let first_node_index = model.node_id0 in
  let rec find node_index =
    if is_leaf_index node_index then lnot node_index
    else
      begin
      let node = !Global.nodes.(node_index) in
      match point_plane_test !Global.planes.(node.nplane_id) point with
       | Front -> find node.front
       | Back -> find node.back
      end in
  find first_node_index
;;



(* ******************
Marque par effet de bord toutes les feuilles considres comme potentiellement
visibles (de par le PVS) de l'endroit o se trouve le joueur.
*)
(* ******************
val find_visible_leaves_from_position : Qheader.vec3_t -> unit
*)
let find_visible_leaves_from_position model point =
  (* Considre aucune feuille visible au dpart *)
  for i = 0 to !Global.numleaves - 1 do
    !Global.visible_leaves.(i) <- false done ;
  (* Recherche l'index de la feuille abritant le joueur *)
  let l_index = find_lying_leaf model point in
  if l_index = 0 || !Global.leaves.(l_index).vislist = -1 then
    begin
    (* Soit on est en dehors d'une feuille (hors du niveau ?), soit la  *)
    (* feuille ne contient pas d'infos de visibilit et dans ces cas on *)
    (* est oblig de considrer que toutes les feuilles sont visibles.  *)
    for i = 0 to !Global.numleaves - 1 do
      !Global.visible_leaves.(i) <- true done
    end
  else
    begin
    (* On va marquer visibles toutes les feuilles considres comme    *)
    (* visibles dans la vislist de la feuille abritant le joueur. On   *)
    (* utilise la routine de dcompression fournie dans les Unofficial *)
    (* Quake Specs, mais en version perso et dbugge...               *)
    let v = ref !Global.leaves.(l_index).vislist in
    let l = ref 1 in    (* Index dans le tableau de visibilit des feuilles *)
    (* Nombre de feuilles du *model* et non global ! Ne prend pas en compte *)
    (* la feuille solide 0, donc il y a 1 feuille de plus moralement.       *)
    let numleaves = model.numleafs in
    while !l <= numleaves do
      if !Global.visilists.[!v] = '\000' then (* Valeur 0: feuil. invisibles *)
        begin
	(* Sauter toutes les feuilles invisibles *)
        l := !l + 8 * Char.code !Global.visilists.[!v + 1] ;
	(* Incrmenter 1 1er coup pour compenser l'octet de longueur *)
        incr v
        end
      else                      (* Tagger 8 feuilles, si ncessaire *)
        begin                   (* Examiner let bits de droite  gauche *)
	let current_byte = Char.code !Global.visilists.[!v] in
	for bitshift = 0 to 7 do
          if current_byte land (1 lsl bitshift) <> 0 then
	    !Global.visible_leaves.(!l) <- true ;        (* Feuille visible *)
          incr l
        done
        end ;
      incr v                (* Octet suivant dans la visilist *)
    done
    end
;;



(* ******************
Propage de bas en haut l'information de visibilit des feuilles (information
infre  partir du PVS) aux noeuds qui contiennent ces feuilles.
*)
(* ******************
val find_visible_nodes : unit -> unit
*)
let find_visible_nodes model =
  let first_node_index = model.node_id0 in
  let rec propagate node_index =
    (* Si la feuille est visible, alors a doit remonter au noeud *)
    if is_leaf_index node_index then !Global.visible_leaves.(lnot node_index)
    else
      begin
      let node = !Global.nodes.(node_index) in
      (* On force le parcours pour viter la laziness du || *)
      let front_vis = propagate node.front in
      let back_vis = propagate node.back in
      let curnode_vis = front_vis || back_vis in
      !Global.visible_nodes.(node_index) <- curnode_vis ;
      curnode_vis
      end in
  (* On se moque du rsultat final *)
  let _ = propagate first_node_index in
  ()
;;



(* ??? Faut-il remettre le tableau de visibilite des faces a 0 ??? *)

(* ******************
Marque chaque face de la feuille comme tant visible
*)
(* ******************
val tag_leaf_faces_as_visible : Qheader.dleaf_t -> unit
*)
let tag_leaf_faces_as_visible leaf =
 let faces_number = leaf.lface_num in
 let first_face = leaf.lface_id in
 for i = 0 to faces_number - 1 do
   (* Attention, 'first_face' reference un index dans la table des  *)
   (* listes de faces, et non directement dans la tables des faces. *)
   let face_id_in_faces_lists = !Global.lfaces.(first_face + i) in
   !Global.visible_faces.(face_id_in_faces_lists) <- true
   done
;;



(* REMARQUE: Faut-il remettre le tableau de visibilite des noeuds a 0 ? *)
(* REPONSE: Non, car je le fais incrmentalement lors de la remonte de *)
(* visibilit bottom-up.                                                *)

(* ******************
Parcours le BSP et regarde les noeud/feuilles marqus visibles tombant dans
le frustum. Si un noeud est marqu mais n'est pas dans le frustum alors on le
d-marque. Si une feuille est marque et dans le frustum, alors on marque
toutes les faces de la feuille comme visible.
*)
(* ******************
val cull_nodes_out_of_frustum : Frustum.frustum -> unit
*)
let cull_nodes_out_of_frustum model frustum =
  let first_node_index = model.node_id0 in
  let rec explore node_index =
    if is_leaf_index node_index then
      begin
      let leaf_index = lnot node_index in
      let leaf = !Global.leaves.(leaf_index) in
      if Frustum.leaf_is_in_frustum frustum leaf then
	tag_leaf_faces_as_visible leaf
      end
    else
      begin
      if !Global.visible_nodes.(node_index) then
        begin
        let node = !Global.nodes.(node_index) in
        if Frustum.node_is_in_frustum frustum node
          then ( explore node.front ; explore node.back)
          else !Global.visible_nodes.(node_index) <- false
        end
      end in
  explore first_node_index
;;



(* ******************
Plan face au joueur en dea duquel on considre que les choses ne sont pas
visibles. En fait, plan de clipping en profondeur afin d'viter des calculs
de projection incohrents.
*)
(* ******************
val clipping_plane : Qheader.plane_t
*)
let clipping_plane =
  { normal = vec3_normalize { x = 1.0 ; y = 0.0 ; z = 0.0 } ;
    dist = -. 10.0 ; ptype = 6 }
;;



(* ******************
Affichage d'un polygone reprsentant une face. Se charge de "fabriquer" le
polygone dans l'espace  partir de la liste de ses vertices. Ensuite,
transforme ce polygone en camera-space (ce qui cre des vertices Glide
d'ailleurs, mme s'ils ont encore une smantique 3D et non 2D comme celle
de Glide). Ensuite, clippe le polygone en profondeur. Puis projette le
rsultat (ce qui donne alors un polygone 2D, avec des vertices Glides prenant
enfin toute leur smantique 2D, avec en plus le passage en repre cran).
Pour plus d'infos sur cette passe, consulter le commentaire de la fonction
'Math.project_polygon3d_in_place'.
Pour finir, on clippe le polygone 2D  la taille de l'cran d'affichage,
on slectionne la texture et on balance le tout  la carte.
*)
(* ******************
val draw_face : Qheader.face_t -> unit
*)
let draw_face face =
  let number_edges = face.ledge_num in
  (* On met  jour par effet de bord les paramtres de textures pour la face *)
  let texinfo = !Global.surfaces.(face.texinfo_id) in
  Camera._vectorS := texinfo.vectorS ;
  Camera._distS := texinfo.distS ;
  Camera._vectorT := texinfo.vectorT ;
  Camera._distT := texinfo.distT ;
  let first_edge = face.ledge_id in
  (* On cre la liste des vertices reprsentant le polygone *)
  (* Et au passage on transforme ses coordonnes en camera-space *)
  let vertices = Array.create number_edges Global.dummy_glide_vertex in
  for i = 0 to number_edges - 1 do
    let edge_id_in_edges_lists = !Global.ledges.(first_edge + i) in
    let vertex =
      if edge_id_in_edges_lists < 0 then
	begin
	(* Il faut prendre l'oppos de l'indice de l'edge *)
	let edge = !Global.edges.(-edge_id_in_edges_lists) in
	!Global.vertices.(edge.vertex1)
	end
      else
	begin
	let edge = !Global.edges.(edge_id_in_edges_lists) in
	!Global.vertices.(edge.vertex0)
	end in
    (* Transformation en camera-space *)
    vertices.(i) <- Camera.vertex_to_camera_space Global.camera vertex
  done ;
  (* Maintenant, on la depth-clip *)
  let clipped_vertices =
    Clip3d.clip_polygon3d_to_3d_plane vertices clipping_plane in
  (* Maintenant on la projette en place et on en fait des vertices 2D *)
  project_polygon3d_in_place clipped_vertices ;
  (* On clippe a la taille de l'cran *)
  let final_polygon = Clip2d.clip_2d_poly_to_screen clipped_vertices in
  (* Et on dessine *)
  let polygon_size = Array.length final_polygon in
  if polygon_size > 0 then
    begin
    let ti = texinfo.texture_id in
    let coord_scale = !Global.texture_scaling.(ti) in
    (* Innefficace, mais c'est une approximation pour le moment *)
    for i = 0 to polygon_size - 1 do
      let v = final_polygon.(i) in
      v.Gr.tmu.(0).Gr.tmu_sow <- v.Gr.tmu.(0).Gr.tmu_sow *. coord_scale ;
      v.Gr.tmu.(0).Gr.tmu_tow <- v.Gr.tmu.(0).Gr.tmu_tow *. coord_scale
    done ;
    (* Slectionne la texture qui va bien *)
    Gr.tex_source 0 !Global.mipaddrs.(ti) Gr.MIPMAPLEVEL_MASK_BOTH
                  !Global.miptexs.(ti) ;
    (* Et hop, on dessine *)
    Gr.draw_planar_polygon final_polygon polygon_size
    end
;;



(* ******************
Rendu des polygones se trouvant  un noeud du BSP. En fait, principalement
c'est une boucle qui appelle la fonction de rendu d'une face pour toutes
les faces prsentes  ce noeud.
*)
(* ******************
val render_node_faces : Qheader.node_t -> Qheader.ushort_t -> unit
*)
let render_node_faces node side =
  let first_face = node.nface_id in
  let number_faces = node.nface_num in
  (* Attention, comme on est dans un node, on va directement chercher *)
  (* la face dans la table des faces et non par la liste d'indirection *)
  for i = 0 to number_faces - 1 do
    if !Global.visible_faces.(first_face + i) then
      begin
      (* La face est marque visible *)
      let face = !Global.faces.(first_face + i) in
      if face.side = side then draw_face face
      end
    done
;;



(* ******************
Rendu spcialis au model 0, c'est--dire au tableau lui-mme.
*)
(* ******************
val render_level : Camera.camera -> unit
*)
let render_level model camera =
  let first_node_index = model.node_id0 in
  let rec rec_render_node node_index =
    if node_index >= 0 && !Global.visible_nodes.(node_index) then
      begin
      let node = !Global.nodes.(node_index) in
      let node_plane = !Global.planes.(node.nplane_id) in
      if point_plane_test node_plane camera.Camera.position = Back then
	begin
	rec_render_node node.front ;
	render_node_faces node 1 ;
	rec_render_node node.back
	end
      else
	begin
	rec_render_node node.back ;
	render_node_faces node 0 ;
	rec_render_node node.front
	end
      end in
  rec_render_node first_node_index
;;



(* ******************
Wrapper qui enchane le prparation du rendu grce aux structures de BSP
et de PVS, puis qui passe  l'affichage proprement dit.
*)
(* ******************
val render : Camera.camera -> unit
*)
let render model camera =
  (* Commencer par marquer toutes les feuilles potentiellement *)
  (* visibles  partir de la position courante *)
  find_visible_leaves_from_position model camera.Camera.position ;
  (* Propager bottom-up l'info de visibilit des feuilles aux noeuds *)
  find_visible_nodes model ;
  let new_frustum = Frustum.transform_frustum Global.frustum camera in
  (* Marquer les faces des feuilles tombant dans le frustum comme visibles *)
  (* au passage, d-marquer les nodes ne tombant pas dans le frustum et    *)
  (* qui auraient ete marqus abusivement au parcours bottom-up prcdent. *)
  cull_nodes_out_of_frustum model new_frustum ;
  render_level model camera ;
  ()
;;
