let screen_width = 640.0 ;;
let screen_height = 480.0 ;;



let inside_bottom (_, y) = y >= 0.0 ;;
let inside_top (_, y) = y < screen_height ;;
let inside_left (x, _) = x >= 0.0 ;;
let inside_right (x, _) = x < screen_width ;;



let intersect_bottom (x1, y1) (x2, y2) =
 let inv_tan_alpha = (x2 -. x1) /. (y2 -. y1) in
 let x = x1 +. (0.0 -. y1) *. inv_tan_alpha in
 let y = 0.0 in
 (x, y)
;;

let intersect_top (x1, y1) (x2, y2) =
 let inv_tan_alpha = (x2 -. x1) /. (y2 -. y1) in
 let x = x2 +. ((screen_height -. 1.0) -. y2) *. inv_tan_alpha in
 let y = (screen_height -. 1.0) in
 (x, y)
;;

let intersect_left (x1, y1) (x2, y2) =
 let tan_alpha = (y2 -. y1) /. (x2 -. x1) in
 let y = y1 +. (0.0 -. x1) *. tan_alpha in
 let x = 0.0 in
 (x, y)
;;

let intersect_right (x1, y1) (x2, y2) =
 let tan_alpha = (y2 -. y1) /. (x2 -. x1) in
 let y = y2 +. ((screen_width -. 1.0) -. x2) *. tan_alpha in
 let x = (screen_width -. 1.0) in
 (x, y)
;;


(* Tableaux de vertices temporaires pour eviter les allocs *)
let vertices_bottom = Array.create 30 (0.0, 0.0) ;;
let vertices_top = Array.create 30 (0.0, 0.0) ;;
let vertices_left = Array.create 30 (0.0, 0.0) ;;
let vertices_right = Array.create 30 (0.0, 0.0) ;;


(* vertices_ : vertices en entree *)
(* numv_ : nombre de vertices en entree *)
(* output_array : tableau ou ranger les vertices en sortie *)
(* Retourne le nombr ede vertices en sortie *)
let clip_poly inside_ intersect_ vertices_ numv_ output_ =
 if numv_ = 0 then 0
 else
 let output_size = ref 0 in
 let s = ref vertices_.(numv_ - 1) in
 for j = 0 to (numv_ - 1) do
  let p = ref vertices_.(j) in
  if inside_ !p then
   if inside_ !s then
     begin
     output_.(!output_size) <- !p ;
     incr output_size
    end
   else
    begin
    let i = intersect_ !s !p in
    output_.(!output_size) <- i ;
    incr output_size ;
    output_.(!output_size) <- !p ;
    incr output_size
    end
  else
   if inside_ !s then
    begin
    let i = intersect_ !s !p in
    output_.(!output_size) <- i ;
    incr output_size
    end ;
  s := !p
 done ;
 !output_size
;;


let clip_2d_poly_to_screen vertices =
 let numv = Array.length vertices in
 let num_bottom =
   clip_poly inside_bottom intersect_bottom vertices numv
             vertices_bottom in
 let num_top =
   clip_poly inside_top intersect_top vertices_bottom num_bottom
             vertices_top in
 let num_left =
   clip_poly inside_left intersect_left vertices_top num_top
             vertices_left in
 let num_right =
   clip_poly inside_right intersect_right vertices_left num_left
             vertices_right in
 let result = Array.make num_right (0, 0) in
 for i = 0 to num_right - 1 do
   let (x, y) = vertices_right.(i) in
   result.(i) <- (truncate x, truncate y)
   done ;
 result
;;
