open Qheader ;;


(* 3D clipping global stuff *)
(* Tableau de vertices temporaires pour eviter les allocs *)
let tmp_vertices =
  Array.init 30 (fun _ -> { vx = 0.0 ; vy = 0.0 ; vz = 0.0 }) ;;
let nb_tmp_vertices = ref 0 ;;
let epsilon = 0.0 ;;


let clip_polygon3d_to_3d_plane polygon plane =
  nb_tmp_vertices := 0 ;
  let allin = ref true and allout = ref true in
  let distances =
    Array.map (fun vertex ->
                let d = Math.vertex_plane_distance plane vertex in
		if d < (-. epsilon) then allin := false ;
		if d >= epsilon then allout := false ;
		d)
              polygon in
  if !allin then polygon
  else if !allout then [||]
  else
    let nb_vertices = Array.length polygon in
    (* Start with the last vertex as first treated vertex *)
    let v1 = ref polygon.(nb_vertices - 1) in
    let dist1 = ref distances.(nb_vertices - 1) in
    let inside = ref (!dist1 >= 0.0) in

    for i = 0 to nb_vertices - 1 do
      let v2 = polygon.(i) in
      let dist2 = distances.(i) in
      if !inside && dist2 >= 0.0 then
	begin
	(* Both in *)
	tmp_vertices.(!nb_tmp_vertices) <- v2 ;
	incr nb_tmp_vertices
	end
      else
	begin
	if (not !inside) && dist2 >= epsilon then
	  begin
	  (* Coming in *)
	  inside := true ;
	  let d = !dist1 /. (!dist1 -. dist2) in
	  let tmpv = {
	    vx = !v1.vx +. (v2.vx -. !v1.vx) *. d ;
	    vy = !v1.vy +. (v2.vy -. !v1.vy) *. d ;
	    vz = !v1.vz +. (v2.vz -. !v1.vz) *. d } in
	    (* Output vertices *)
	    tmp_vertices.(!nb_tmp_vertices) <- tmpv ;
	    incr nb_tmp_vertices ;
            tmp_vertices.(!nb_tmp_vertices) <- v2 ;
	    incr nb_tmp_vertices
	   end
	else
	  begin
	  if !inside && dist2 < (-. epsilon) then
	    begin
	    (* Going out *)
	    inside := false ;
	    let d = !dist1 /. (!dist1 -. dist2) in
	    let tmpv = {
	      vx = !v1.vx +. (v2.vx -. !v1.vx) *. d ;
	      vy = !v1.vy +. (v2.vy -. !v1.vy) *. d ;
	      vz = !v1.vz +. (v2.vz -. !v1.vz) *. d } in
	     (* Output vertex *)
	     tmp_vertices.(!nb_tmp_vertices) <- tmpv ;
	     incr nb_tmp_vertices
	     end
	  else
	    () (* Both out *)
	  end
	end ;
      v1 := polygon.(i) ;
      dist1 := distances.(i)
    done ;
    let new_vertices = Array.sub tmp_vertices 0 !nb_tmp_vertices in
    new_vertices
;;
