open Math ;;
open Camera ;;

let start_time = ref 0.0 ;;
let end_time = ref 0.0 ;;
let frame_count = ref 0 ;;

let old_camera_position =
  { Qheader.x = 544.0; Qheader.y = 288.0 ; Qheader.z = 32.0 } ;;


let restore_old_camera_position () =
  Global.camera.position.Qheader.x <- old_camera_position.Qheader.x ;
  Global.camera.position.Qheader.y <- old_camera_position.Qheader.y ;
  Global.camera.position.Qheader.z <- old_camera_position.Qheader.z
;;


let update_camera_old_position () =
  old_camera_position.Qheader.x <- Global.camera.position.Qheader.x ;
  old_camera_position.Qheader.y <- Global.camera.position.Qheader.y ;
  old_camera_position.Qheader.z <- Global.camera.position.Qheader.z
;;



let find_initial_camera_position () =
  try
    for i = 0 to !Global.numentities - 1 do
      match !Global.entities.(i) with
       | Qheader.Info_player_start position ->
	   Global.camera.position.Qheader.x <- position.Qheader.x ;
	   Global.camera.position.Qheader.y <- position.Qheader.y ;
	   Global.camera.position.Qheader.z <- position.Qheader.z ;
	   update_camera_old_position () ;
	   raise End_of_file
       | _ -> ()
    done ;
    Printf.printf "Warning not start position found\n"
  with End_of_file -> ()
;;



Graphics.open_graph ":0.0 640x480+0+0" ;;

Arg.parse [] (fun filename -> Readbspfile.load_file filename) "Bla bla" ;;
let tmp_matrix = Matrix.identity_matrix () ;;
find_initial_camera_position () ;;
start_time := Unix.gettimeofday () ;;
frame_count := 1 ;;
Render.render Global.camera ;;
while true do
  if Graphics.key_pressed () then
    begin
    (match Graphics.read_key () with
     | '8' ->  (* Up arrow : move forward *)
	 let tmp = { Qheader.x = 30.0 ; Qheader.y = 0.0; Qheader.z = 0.0 } in
	 let tmp' = vec3_rotate tmp Global.camera.rotation in
	 Global.camera.position.Qheader.x <-
           Global.camera.position.Qheader.x +. tmp'.Qheader.x ;
	 Global.camera.position.Qheader.y <-
	   Global.camera.position.Qheader.y +. tmp'.Qheader.y ;
	 Global.camera.position.Qheader.z <-
           Global.camera.position.Qheader.z +. tmp'.Qheader.z
     | '2' -> (* Down arrow : move backward *)
	 let tmp = { Qheader.x = -.30.0 ; Qheader.y = 0.0; Qheader.z = 0.0 } in
	 let tmp' = vec3_rotate tmp Global.camera.rotation in
	 Global.camera.position.Qheader.x <-
           Global.camera.position.Qheader.x +. tmp'.Qheader.x ;
	 Global.camera.position.Qheader.y <-
	   Global.camera.position.Qheader.y +. tmp'.Qheader.y ;
	 Global.camera.position.Qheader.z <-
           Global.camera.position.Qheader.z +. tmp'.Qheader.z
     | '-' -> (* Minus : go down *)
	 let tmp = { Qheader.x = 0.0 ; Qheader.y = 0.0; Qheader.z = -30.0 } in
	 let tmp' = vec3_rotate tmp Global.camera.rotation in
	 Global.camera.position.Qheader.x <-
           Global.camera.position.Qheader.x +. tmp'.Qheader.x ;
	 Global.camera.position.Qheader.y <-
	   Global.camera.position.Qheader.y +. tmp'.Qheader.y ;
	 Global.camera.position.Qheader.z <-
           Global.camera.position.Qheader.z +. tmp'.Qheader.z
     | '+' -> (* Minus : go up *)
	 let tmp = { Qheader.x = 0.0 ; Qheader.y = 0.0; Qheader.z = 30.0 } in
	 let tmp' = vec3_rotate tmp Global.camera.rotation in
	 Global.camera.position.Qheader.x <-
           Global.camera.position.Qheader.x +. tmp'.Qheader.x ;
	 Global.camera.position.Qheader.y <-
	   Global.camera.position.Qheader.y +. tmp'.Qheader.y ;
	 Global.camera.position.Qheader.z <-
           Global.camera.position.Qheader.z +. tmp'.Qheader.z

     | '1' -> (* 1 : strafe left *)
	 let tmp = { Qheader.x = 0.0 ; Qheader.y = 30.0; Qheader.z = 0.0 } in
	 let tmp' = vec3_rotate tmp Global.camera.rotation in
	 Global.camera.position.Qheader.x <-
           Global.camera.position.Qheader.x +. tmp'.Qheader.x ;
	 Global.camera.position.Qheader.y <-
	   Global.camera.position.Qheader.y +. tmp'.Qheader.y ;
	 Global.camera.position.Qheader.z <-
           Global.camera.position.Qheader.z +. tmp'.Qheader.z
     | '3' -> (* 3 : strafe right *)
	 let tmp = { Qheader.x = 0.0 ; Qheader.y = -.30.0; Qheader.z = 0.0 } in
	 let tmp' = vec3_rotate tmp Global.camera.rotation in
	 Global.camera.position.Qheader.x <-
           Global.camera.position.Qheader.x +. tmp'.Qheader.x ;
	 Global.camera.position.Qheader.y <-
	   Global.camera.position.Qheader.y +. tmp'.Qheader.y ;
	 Global.camera.position.Qheader.z <-
           Global.camera.position.Qheader.z +. tmp'.Qheader.z
     | '6' -> (* Right arrow : turn right *)
	 Matrix.fill_rz_matrix (rad (-. 10.0)) tmp_matrix ;
	 Matrix.mult_matrices Global.camera.rotation tmp_matrix
	                      Global.camera.rotation ;
	 Matrix.fill_rz_matrix (rad (10.0)) tmp_matrix ;
	 Matrix.mult_matrices tmp_matrix Global.camera.irotation
	                      Global.camera.irotation
     | '4' -> (* Left arrow : turn left *)
	 Matrix.fill_rz_matrix (rad 10.0) tmp_matrix ;
	 Matrix.mult_matrices Global.camera.rotation tmp_matrix
	                      Global.camera.rotation ;
	 Matrix.fill_rz_matrix (rad (-. 10.0)) tmp_matrix ;
	 Matrix.mult_matrices tmp_matrix Global.camera.irotation
	                      Global.camera.irotation
     | '7' ->
	 (* Roll left *)
	 Matrix.fill_rx_matrix (rad (-. 10.0)) tmp_matrix ;
	 Matrix.mult_matrices Global.camera.rotation tmp_matrix
	                      Global.camera.rotation ;
	 Matrix.fill_rx_matrix (rad 10.0) tmp_matrix ;
	 Matrix.mult_matrices tmp_matrix Global.camera.irotation
	                      Global.camera.irotation
     | '9' ->
	 (* Roll right *)
	 Matrix.fill_rx_matrix (rad 10.0) tmp_matrix ;
	 Matrix.mult_matrices Global.camera.rotation tmp_matrix
	                      Global.camera.rotation ;
	 Matrix.fill_rx_matrix (rad (-. 10.0)) tmp_matrix ;
	 Matrix.mult_matrices tmp_matrix Global.camera.irotation
	                      Global.camera.irotation
     | 'c' -> (* c : clipping on / off *)
	 Global.coldet_active := not !Global.coldet_active
     | '\027' -> (* Escape : quit *)
	 end_time := Unix.gettimeofday () ;
	 Printf.printf "Frame rate : %f f/s\n"
	               (float !frame_count /. (!end_time -. !start_time)) ;
	 exit 0
     | _ -> ()) ;
    if !Global.coldet_active &&
       Collision.exists_collision
	            old_camera_position Global.camera.position
	            !Global.models.(0).Qheader.node_id1
    then restore_old_camera_position ()
    else update_camera_old_position () ;
    Graphics.clear_graph () ;
    Render.render Global.camera ;
    incr frame_count
    end
done
;;
