open Qheader ;;
open Matrix ;;


let pi = 4.0 *. atan 1.0 ;;
let rad x = x *. pi /. 180.0 ;;


(* Produit vectoriel *)
let vec3_crossprod { x = a1 ; y = a2 ; z = a3 } { x = b1 ; y = b2 ; z = b3 } =
  { x = a2 *. b3 -. a3 *. b2 ;
    y = a3 *. b1 -. a1 *. b3 ;
    z = a1 *. b2 -. a2 *. b1 }
;;


(* Produit scalaire *)
let vec3_dotprod { x = a1 ; y = a2 ; z = a3 } { x = b1 ; y = b2 ; z = b3 } =
  a1 *. b1 +. a2 *. b2 +. a3 *. b3
;;


(* Vector normalisation *)
let vec3_normalize { x = a1 ; y = a2 ; z = a3 } =
  let len = sqrt (a1 *. a1 +. a2 *. a2 +. a3 *. a3) in
  { x = a1 /. len ; y = a2 /. len ; z = a3 /. len }
;;


(* Vector length *)
let vec3_length { x = x ; y = y ; z = z } =
  sqrt (x *. x +. y *. y +. z *. z)
;;


(* Rotation de vertex *)
let vertex_rotate { vx = vx ; vy = vy ; vz = vz } matrix =
  let vx' = vx *. matrix.m11 +. vy *. matrix.m12 +. vz *. matrix.m13 in
  let vy' = vx *. matrix.m21 +. vy *. matrix.m22 +. vz *. matrix.m23 in
  let vz' = vx *. matrix.m31 +. vy *. matrix.m32 +. vz *. matrix.m33 in
  { vx = vx' ; vy = vy' ; vz = vz' }
;;


(* Rotation de vecteur *)
let vec3_rotate { x = x ; y = y ; z = z } matrix =
  let x' = x *. matrix.m11 +. y *. matrix.m12 +. z *. matrix.m13 in
  let y' = x *. matrix.m21 +. y *. matrix.m22 +. z *. matrix.m23 in
  let z' = x *. matrix.m31 +. y *. matrix.m32 +. z *. matrix.m33 in
  { x = x' ; y = y' ; z = z' }
;;


(* Calcule la distance d'un vertex a` un plan *)
(* Dans l'equation d'un plan, si le vecteur normal est normalise, alors *)
(* le coefficient d = distance du plan au point 0. C'est a dire que si  *)
(* le plan est devant l'origine, la distance sera negative.             *)
let vertex_plane_distance plane { vx = x ; vy = y ; vz = z } =
  let num = plane.normal.x *. x +.
            plane.normal.y *. y +.
            plane.normal.z *. z +.
            plane.dist in
  (* Because normal vector of the plane is assumed to be normalized   *)
  (* we can avoid dividing by the numerator which is hence always 1.0 *)
  (* let denom = sqrt (plane.normal.x *. plane.normal.x +.
		    plane.normal.y *. plane.normal.y +.
		    plane.normal.z *. plane.normal.z) in *)
  num (* / denom *)
;;


type location = Front | Back ;;


(* Teste si un point est devant ou derrie`re un plan *)
let point_plane_test plane { x = x ; y = y ; z = z } =
  (* Because normal vector of the plane is assumed to be *)
  (* normalized we can apply the same optimization than above. *)
  let tmp =
    plane.normal.x *. x +. plane.normal.y *. y +. plane.normal.z *. z in
    if tmp < plane.dist then Back else Front
;;


let project_polygon3d polygon =
  Array.map
     (fun { vx = x ; vy = y ; vz = z } ->
       (* Recentrer sur l'ecran + compenser le fait que l'axe *)
       (* des x-ecran varient dans le sens inverse de l'axe   *)
       (* des y dans le monde "reel".                         *)
       (* On multiplie par 320 = 640/2 pour le fov.           *)
       ((320.0 -. 320.0 *. y /. x),
	(240.0 +. 320.0 *. z /. x)))
     polygon
;;
