type matrix3x3 = {
  mutable m11 : float ;
  mutable m12 : float ;
  mutable m13 : float ;
  mutable m21 : float ;
  mutable m22 : float ;
  mutable m23 : float ;
  mutable m31 : float ;
  mutable m32 : float ;
  mutable m33 : float
  } ;;


let identity_matrix () = {
  m11 = 1.0 ; m12 = 0.0 ; m13 = 0.0 ;
  m21 = 0.0 ; m22 = 1.0 ; m23 = 0.0 ;
  m31 = 0.0 ; m32 = 0.0 ; m33 = 1.0 }
;;


(*
Sens de rotation directe: Mcamera *. Mrotation_directe
Sens de rotation inverse: Mrotation_inverse * Mcamera
Mrotation_inverse = transposee de Mrotation_directe
*)

let mult_matrices left right result =
  let m11l = left.m11 in let m11r = right.m11 in
  let m12l = left.m12 in let m12r = right.m12 in
  let m13l = left.m13 in let m13r = right.m13 in
  let m21l = left.m21 in let m21r = right.m21 in
  let m22l = left.m22 in let m22r = right.m22 in
  let m23l = left.m23 in let m23r = right.m23 in
  let m31l = left.m31 in let m31r = right.m31 in
  let m32l = left.m32 in let m32r = right.m32 in
  let m33l = left.m33 in let m33r = right.m33 in
  result.m11 <- m11l *. m11r +. m12l *. m21r +. m13l *. m31r ;
  result.m12 <- m11l *. m12r +. m12l *. m22r +. m13l *. m32r ;
  result.m13 <- m11l *. m13r +. m12l *. m23r +. m13l *. m33r ;
  result.m21 <- m21l *. m11r +. m22l *. m21r +. m23l *. m31r ;
  result.m22 <- m21l *. m12r +. m22l *. m22r +. m23l *. m32r ;
  result.m23 <- m21l *. m13r +. m22l *. m23r +. m23l *. m33r ;
  result.m31 <- m31l *. m11r +. m32l *. m21r +. m33l *. m31r ;
  result.m32 <- m31l *. m12r +. m32l *. m22r +. m33l *. m32r ;
  result.m33 <- m31l *. m13r +. m32l *. m23r +. m33l *. m33r
;;


let fill_ry_matrix a matrix =
  matrix.m11 <- cos a ;
  matrix.m12 <- 0.0 ;
  matrix.m13 <- sin a ;
  matrix.m21 <- 0.0 ;
  matrix.m22 <- 1.0 ;
  matrix.m23 <- 0.0 ;
  matrix.m31 <- -. sin a ;
  matrix.m32 <- 0.0 ;
  matrix.m33 <- cos a
;;


let fill_rz_matrix a matrix =
  matrix.m11 <- cos a ;
  matrix.m12 <- -. sin a ;
  matrix.m13 <- 0.0 ;
  matrix.m21 <- sin a ;
  matrix.m22 <- cos a ;
  matrix.m23 <- 0.0 ;
  matrix.m31 <- 0.0 ;
  matrix.m32 <- 0.0 ;
  matrix.m33 <- 1.0
;;


let fill_rx_matrix a matrix =
  matrix.m11 <- 1.0 ;
  matrix.m12 <- 0.0 ;
  matrix.m13 <- 0.0 ;
  matrix.m21 <- 0.0 ;
  matrix.m22 <- cos a ;
  matrix.m23 <- -. sin a ;
  matrix.m31 <- 0.0 ;
  matrix.m32 <- sin a ;
  matrix.m33 <- cos a
;;
