open Qheader ;;
open Math ;;



(* let walk_bsp_tree_fb node_index =
   if player is front then front / back
   else back / front
;;

let walk_bsp_tree_bf node_index =
  if player is front then back / front
  else front / back
;; *)


(* Teste si un index est un index de feuille *)
let is_leaf_index index = index land 0x8000 <> 0 ;;
(* Teste si un index est un index de noeud *)
let is_node_index index = index land 0x8000 = 0 ;;



(* Recherche dans quelle feuille du BSP se trouve le point argument *)
let find_lying_leaf point =
  let world_model = !Global.models.(0) in
  let first_node_index = world_model.node_id0 in
  let rec find node_index =
    if is_leaf_index node_index then lnot node_index
    else
      begin
      let node = !Global.nodes.(node_index) in
      match point_plane_test !Global.planes.(node.nplane_id) point with
       | Front -> find node.front
       | Back -> find node.back
      end in
  find first_node_index
;;



(* Marque toutes les feuilles conside're'es comme *)
(* visibles de l'endroit ou` se trouve le joueur. *)
let find_visible_leaves_from_position point =
  (* Conside`re aucune feuille visible au depart *)
  for i = 0 to !Global.numleaves - 1 do
    !Global.visible_leaves.(i) <- false done ;
  (* Recherche l'index de la feuille abritant le joueur *)
  let l_index = find_lying_leaf point in
  if l_index = 0 || !Global.leaves.(l_index).vislist = -1 then
    begin
    (* Soit on est en dehors d'une feuille (hors du niveau ?), soit la   *)
    (* feuille ne contient pas d'infos de visibilite' et dans ces cas on *)
    (* est oblige' de conside'rer que toutes les feuilles sont visibles  *)
    for i = 0 to !Global.numleaves - 1 do
      !Global.visible_leaves.(i) <- true done
    end
  else
    begin
    (* On va marquer visibles toutes les feuilles considres comme    *)
    (* visibles dans la vislist de la feuille abritant le joueur. On   *)
    (* utilise la routine de dcompression fournie dans les Unofficial *)
    (* Quake Specs, mais en version perso dbugge...                  *)
    let v = ref !Global.leaves.(l_index).vislist in
    let l = ref 1 in    (* Index dans le tableau de visibilite des feuilles *)
    (* Nombre de feuilles du *model* et non global ! Ne prend pas en compte *)
    (* la feuille solide 0, donc y'a 1 feuille de plus moralement *)
    let numleaves = !Global.models.(0).numleafs in
    while !l <= numleaves do
      if !Global.visilists.[!v] = '\000' then (* Valeur 0: feuil. invisibles *)
        begin
	(* Sauter toutes les feuilles invisibles *)
        l := !l + 8 * Char.code !Global.visilists.[!v + 1] ;
	(* Incrmenter 1 1er coup pour compenser l'octet de longueur *)
        incr v
        end
      else                      (* Tagger 8 feuilles, si necessaire *)
        begin                   (* Examiner let bits de droite  gauche *)
	let current_byte = Char.code !Global.visilists.[!v] in
	for bitshift = 0 to 7 do
          if current_byte land (1 lsl bitshift) <> 0 then
	    !Global.visible_leaves.(!l) <- true ;        (* Feuille visible *)
          incr l
        done
        end ;
      incr v                (* Octet suivant dans la visilist *)
    done
    end
;;



(* Propage de bas en haut l'information de visibilit des *)
(* feuilles aux noeuds qui les contiennent                *)
let find_visible_nodes () =
  let world_model = !Global.models.(0) in
  let first_node_index = world_model.node_id0 in
  let rec propagate node_index =
    (* Si la feuille est visible, alors a doit remonter au noeud *)
    if is_leaf_index node_index then !Global.visible_leaves.(lnot node_index)
    else
      begin
      let node = !Global.nodes.(node_index) in
      (* On force le parcours pour viter la laziness du || *)
      let front_vis = propagate node.front in
      let back_vis = propagate node.back in
      let curnode_vis = front_vis || back_vis in
      !Global.visible_nodes.(node_index) <- curnode_vis ;
      curnode_vis
      end in
  (* On se moque du rsultat final *)
  let _ = propagate first_node_index in
  ()
;;

(* ??? Faut-il remettre le tableau de visibilite des faces a 0 ??? *)

(* Marque chaque face de la feuille comme etant visible *)
let tag_leaf_faces_as_visible leaf =
 let faces_number = leaf.lface_num in
 let first_face = leaf.lface_id in
 for i = 0 to faces_number - 1 do
   (* Attention, 'first_face' reference un index dans la table des  *)
   (* listes de faces, et non directement dans la tables des faces. *)
   let face_id_in_faces_lists = !Global.lfaces.(first_face + i) in
   !Global.visible_faces.(face_id_in_faces_lists) <- true
   done
;;



(* ??? Faut-il remettre le tableau de visibilite des noeuds a 0 ?     *)
(* Non, je le fais incrmentalement lors de la remonte de visibilit *)
(* bottom-up.                                                         *)

(* Parcours le BSP et regarde les noeud/feuilles marqus visibles tombant *)
(* dans le frustum. Si un noeud est marqu mais n'est pas dans le frustum *)
(* alors on le d-marque. Si une feuille est marque et dans le frustum,  *)
(* alors on marque toutes les faces de la feuille comme visible.          *)
let cull_nodes_out_of_frustum frustum =
  let world_model = !Global.models.(0) in
  let first_node_index = world_model.node_id0 in
  let rec explore node_index =
    if is_leaf_index node_index then
      begin
      let leaf_index = lnot node_index in
      let leaf = !Global.leaves.(leaf_index) in
      if Frustum.leaf_is_in_frustum frustum leaf then
	tag_leaf_faces_as_visible leaf
      end
    else
      begin
      if !Global.visible_nodes.(node_index) then
        begin
        let node = !Global.nodes.(node_index) in
        if Frustum.node_is_in_frustum frustum node
          then ( explore node.front ; explore node.back)
          else !Global.visible_nodes.(node_index) <- false
        end
      end in
  explore first_node_index
;;



let clipping_plane =
  { normal = vec3_normalize { x = 1.0 ; y = 0.0 ; z = 0.0 } ;
    dist = -. 10.0 ; ptype = 6 }
;;



let draw_face face =
  let number_edges = face.ledge_num in
  let first_edge = face.ledge_id in
  (* On cre la liste des vertices reprsentant le polygone *)
  (* Et au passage on transforme ses coordonnes en camera-space *)
  let vertices = Array.create number_edges !Global.vertices.(0) in
  for i = 0 to number_edges - 1 do
    let edge_id_in_edges_lists = !Global.ledges.(first_edge + i) in
    let vertex =
      if edge_id_in_edges_lists < 0 then
	begin
	(* Il faut prendre l'oppos de l'indice de l'edge *)
	let edge = !Global.edges.(-edge_id_in_edges_lists) in
	!Global.vertices.(edge.vertex1)
	end
      else
	begin
	let edge = !Global.edges.(edge_id_in_edges_lists) in
	!Global.vertices.(edge.vertex0)
	end in
    (* Trsnaformation en camera-space *)
    vertices.(i) <- Camera.vertex_to_camera_space Global.camera vertex
  done ;
  (* Maintenant, on la depth-clip *)
  let clipped_vertices =
    Clip3d.clip_polygon3d_to_3d_plane vertices clipping_plane in
  (* Maintenant on la projette *)
  let points = project_polygon3d clipped_vertices in
  (* On clippe a la taille de l'cran *)
  let points_clipped = Clip2d.clip_2d_poly_to_screen points in
  (* Et on dessine *)
  Graphics.set_color !Global.faces_color.(face.texinfo_id) ;
  Graphics.fill_poly points_clipped
;;


let render_node_faces node side =
  let first_face = node.nface_id in
  let number_faces = node.nface_num in
  (* Attention, comme on est dans un node, on va directement chercher *)
  (* la face dans la table des faces et non par la liste d'indirection *)
  for i = 0 to number_faces - 1 do
    if !Global.visible_faces.(first_face + i) then
      begin
      (* La face est marque visible *)
      let face = !Global.faces.(first_face + i) in
      if face.side = side then draw_face face
      end
    done
;;



let render_level camera =
  let world_model = !Global.models.(0) in
  let first_node_index = world_model.node_id0 in
  let rec rec_render_node node_index =
    if node_index >= 0 && !Global.visible_nodes.(node_index) then
      begin
      let node = !Global.nodes.(node_index) in
      let node_plane = !Global.planes.(node.nplane_id) in
      if point_plane_test node_plane camera.Camera.position = Back then
	begin
	rec_render_node node.front ;
	render_node_faces node 1 ;
	rec_render_node node.back
	end
      else
	begin
	rec_render_node node.back ;
	render_node_faces node 0 ;
	rec_render_node node.front
	end
      end in
  rec_render_node first_node_index
;;



let render camera =
  (* Commencer par marquer toutes les feuilles potentiellement *)
  (* visibles  partir de la position courante *)
  find_visible_leaves_from_position camera.Camera.position ;
  (* Propager bottom-up l'info de visibilit des feuilles aux noeuds *)
  find_visible_nodes () ;
  let new_frustum = Frustum.transform_frustum Global.frustum camera in
  (* Marquer les faces des feuilles tombant dans le frustum comme visibles *)
  (* au passage, d-marquer les nodes ne tombant pas dans le frustum et    *)
  (* qui auraient ete marqus abusivement au parcours bottom-up prcdent. *)
  cull_nodes_out_of_frustum new_frustum ;
  render_level camera ;
  ()
;;
